<?php

/**
 * Program: Template_ Compiler
 * Version: 2.1.4
 * Date: 2003-12-15
 * Author: Hyeong-Gil Park
 * Email: xtac@xtac.net
 * Homepage: http://www.xtac.net
 * Licence: Freeware.
 * 
 * Idea of "PHP document templating system" is from "FastTemplate(http://www.thewebmasters.net)".
 * Idea of "compiling PHP template and template plugin" is from "Smarty(http://smarty.php.net)".
 * Idea of "caching" is from "Smarty, Pear Cache(http://pear.php.net), and Cached FastTemplate(http://zoned.net:81/~xkahn/php/fasttemplate)".
 * Special thanks to
 *   Seung-Min Kwon,
 *   Jun-Sung Lee,
 *   Jin-Wook Cho,
 *   Soo-Kyeong Hong,
 *   Yo-Han Kim,
 *   Weon-Soon Lee,
 *   Jae-Gyun Yu,
 *   Sang-Wook Kang,
 *   Jae-Sik Kim,
 *   Myung-Soo kim, and
 *   Neotec(Ltd)
 * for good suggestion and feedback.
*/

class Template_Compiler_
{
	function _compile_template(&$tpl, $tpl_path, $cpl_base, $cpl_head)
	{
		$this->compile_dir =$tpl->compile_dir;
		$this->compile_ext =$tpl->compile_ext;
		$this->skin        =$tpl->skin;
		$this->template_dir=$tpl->template_dir;
		$this->permission  =$tpl->permission or $this->permission = 0777;
		$this->prefilter   =$tpl->prefilter;
		$this->postfilter  =$tpl->postfilter;
		$this->safe_mode   =$tpl->safe_mode;
		$this->plugin_fld  =$tpl->plugin_fld;
		$this->tpl_path    =$tpl_path;
		$this->auto_globals=array('_SERVER','_ENV','_COOKIE','_GET','_POST','_FILES','_REQUEST','_SESSION');
		$this->rsv_words   =array('index_', 'size_', 'key_', 'value_');
		$this->func_list   =array(''=>array());
		$this->key_words   =array('true','false','null');
		$this->constants   =array_keys(get_defined_constants());
		$this->functions   =get_defined_functions();
		$this->functions   =array_merge($this->functions['internal'], array('isset','empty','eval','list','include','require','include_once','require_once'));
		$this->safe_mode_functions=array();
		$this->on_ms       =substr(__file__,0,1)!=='/';
	//
		$cpl_ext =trim($this->compile_ext) or $cpl_ext='php';
		$cpl_path=$cpl_base.'.'.$cpl_ext;
		if ($this->on_ms) $cpl_path=strtr($cpl_path, array('\\\\'=>'/','\\'=>'/'));
		if (!@is_file($cpl_path)) {
			$dir=explode('/',$cpl_path);
			$path=$dir[0];
			for ($nodir=0,$i=1,$s=count($dir)-1; $i<$s; $i++) {
				$path.='/'.$dir[$i];
				if ($nodir or @!is_dir($path) and $nodir=1) {
					if (false===@mkdir($path, $this->permission)) {
						$this->report('Error #1', 'cannot create compile directory <b>'.$path.'</b>. check compile directory or permission');
						exit;
					}
					$this->on_ms || @chmod($path, $this->permission);
				}
			}
		}
	//
		$fp_tpl=fopen($tpl_path,'rb');
		$source=fread($fp_tpl,filesize($tpl_path));
		fclose($fp_tpl);

		$nl_cnt=1;
		$nl_del_sum=0;
		$this->nl_del[0]=0;
		$match=array();
		$nl=preg_match('/\r\n|\n|\r/', $source, $match) ? $match[0] : "\r\n";

		$escape_q1=array('\\\\'=>'\\', "\\'"=>"'", '\\n'=>$nl, '\\t'=>"\t", '\\>'=>'>');
		$escape_q2=array('\\\\'=>'\\', '\\"'=>'"', '\\n'=>$nl, '\\t'=>"\t", '\\>'=>'>');
		$split=preg_split('/(<!--{\*|\*}-->|{\*|\*})/', $source, -1, PREG_SPLIT_DELIM_CAPTURE);
		for ($j=0,$i=0,$s=count($split); $i<$s; $i++) {
			if (!($i%2)) {
				$nl_cnt+=substr_count($split[$i], $nl);
				continue;
			}
			switch ($split[$i]) {
			case'<!--{*':
			case    '{*':
				if (substr($split[$i+1],0,1)=='\\') $split[$i+1]=substr($split[$i+1],1);
				elseif (!$j) $j=$i;
				break;
			case '*}-->':
			case '*}'   :
				if (substr($split[$i-1],-1)==='\\')
					$split[$i-1]=substr($split[$i-1],0,-1);
				elseif ($j) {
					if ($j===1 && !trim($split[0])) {
						for ($def_area='',$k=2; $k<$i; $k++) $def_area.=$split[$k];
						preg_match_all('/(?:(?:^|\r\n|\n|\r)[ \t]*)\#(prefilter|postfilter)[ \t]+
							((?:"(?:\\\\"|[^"])*")|(?:\'(?:\\\\\'|[^\'])*\')|(?:[^"\']\S*))
						/x', $def_area, $match, PREG_PATTERN_ORDER);
						for ($k=0,$t=count($match[0]); $k<$t; $k++) {
							if ($this->safe_mode) $match[2][$k]=preg_replace('/<\?(php)/i', '&lt;?$1', $match[2][$k]);
							if ($match[2][$k][0]==="'") $f_string=strtr(substr($match[2][$k],1,-1), $escape_q1);
							elseif ($match[2][$k][0]==='"') $f_string=strtr(substr($match[2][$k],1,-1), $escape_q2);
							else $f_string=$match[2][$k];
							if (!trim($f_string)) {
								$this->$match[1][$k]='';
							} else {
								$f_split=preg_split('/(?<!\\\\)\|/', $f_string);
								if (!trim($f_split[0])) $this->$match[1][$k].=$f_string;
								elseif (!trim($f_split[count($f_split)-1])) $this->$match[1][$k]=$f_string.$this->$match[1][$k];
								else $this->$match[1][$k]=$f_string;
							}
						}
						preg_match_all('/(?:(?:^|\r\n|\n|\r)[ \t]*)\#define[ \t]+
							((?:"(?:\\\\"|[^"])*")|(?:\'(?:\\\\\'|[^\'])*\')|(?:[^"\']\S*))[ \t]+
							((?:"(?:\\\\"|[^"])*")|(?:\'(?:\\\\\'|[^\'])*\')|(?:[^"\']\S*))
						/x', $def_area, $match, PREG_PATTERN_ORDER);
						for ($k=0,$t=count($match[0]); $k<$t; $k++) {
							if ($match[1][$k][0]==="'") $key = strtr(substr($match[1][$k],1,-1), $escape_q1);
							elseif ($match[1][$k][0]==='"') $key = strtr(substr($match[1][$k],1,-1), $escape_q2);
							else $key=$match[1][$k];
							if ($match[2][$k][0]==="'") $val = strtr(substr($match[2][$k],1,-1), $escape_q1);
							elseif ($match[2][$k][0]==='"') $val = strtr(substr($match[2][$k],1,-1), $escape_q2);
							else $val=$match[2][$k];
							$macro[$key]=$val;
						}
					}
					for ($nl_sub_cnt=0,$k=$j; $k<=$i; $k++) {
						$nl_sub_cnt += substr_count($split[$k], $nl);
						$split[$k]='';
					}
					$split[$j-1]=preg_replace('/(^|\r\n|\n|\r)[ \t]*$/', '$1', $split[$j-1]);
					if (preg_match('/^[ \t]*(\r\n|\n|\r)/', $split[$i+1])) {
						$nl_del_sum++;
						$split[$i+1]=preg_replace('/^[ \t]*(\r\n|\n|\r)/', '', $split[$i+1]);
					}
					$nl_del_sum += $nl_sub_cnt;
					$nl_cnt -= $nl_sub_cnt;
					$this->nl_del[$nl_cnt] = $nl_del_sum;
					$j=0;
				}
			}
		}
		krsort($this->nl_del);
		$source=implode('',$split);
	//
		if (!empty($macro)) $source=strtr($source, $macro);
	//
		if ($this->safe_mode) {
			$file=dirname(__File__).'/'.$this->plugin_fld.'/functions.safe_mode.ini';
			if (@is_file($file)) {
				$fp=fopen($file, 'r');
				$fc=fread($fp, filesize($file));
				fclose($fp);
				$this->safe_mode_functions=preg_split('/\s+/', trim(strtolower(preg_replace('/;[^\n\r]*/','',$fc))));
			} else {
				$this->report('Warning #4', 'safe mode : cannot find safe function list file <b>'.$file.'</b>', true, false);
			}
			if (ini_get('short_open_tag')) $safe_map['/<\?/']='&lt;?';
			else $safe_map['/<\?(php)/i']='&lt;?$1';
			if (ini_get('asp_tags')) $safe_map['/<%/']='&lt;?';
			$source=preg_replace(array_keys($safe_map),array_values($safe_map),$source);
		}
	//
		if (trim($this->prefilter)) $source=$this->_filter($source, 'pre');
	//
		$this->_control_stack=array();
		$this->_loop_depth=0;
		$this->_loop_stack=array();
		$this->_loop_trace=array();
		$this->_size_trace=array();
		$this->_size_prefix='';
		$this->in_div ='';
		$this->nl_cnt = 1;
	//
		$division=array();
		$divnames=array();
		$nl = preg_match('/\r\n|\n|\r/', $source, $match) ? $match[0] : "\r\n";
		$this->_split=preg_split('/(<!--{(?!`)|(?<!`)}-->|{(?!`)|(?<!`)})/i', $source, -1, PREG_SPLIT_DELIM_CAPTURE);
		for ($mark_tpl=0,$this->_index=0,$s=count($this->_split); $this->_index<$s; $this->_index++) {
			if (!($this->_index % 2)) {
				$this->nl_cnt += substr_count($this->_split[$this->_index], $nl);
				continue;
			}
			switch ($this->_split[$this->_index]) {
			case'<!--{':
			case    '{':
				$mark_tpl=$this->_index;
				break;
			case '}-->':
			case '}'   :
				if ($mark_tpl!==$this->_index-2) break;
				if (!$object=$this->_compile_statement($this->_split[$this->_index-1])) break;
				if (is_array($object)) {
					if ($object[0]===8) { // 1:echo; 2:loop,if,else,elseif,end; 4:include; 8:div; 16:comment;
						if ($object[1]) {
							if (in_array($object[1], $divnames)) {
								$this->report('Error #2', 'template division id <b>'.$object[1].'</b> is overlapped', true, true);
								exit;
							}
							$divnames[]=$object[1];
							$num=count($division);
							$division[$num]=array('name'=>$object[1],'start'=>$this->_index-1);
							if ($num && !isset($division[--$num]['end'])) $division[$num]['end']=$this->_index-1;
							$this->in_div=$object[1];
							$this->func_list[$object[1]]=array();
						} elseif ($num=count($division) and !isset($division[--$num]['end'])) {
							$division[$num]['end']=$this->_index-1;
							$this->in_div='';
						}
						$this->_split[$mark_tpl-1] = preg_replace('/\s*$/',  '', $this->_split[$mark_tpl-1]);
						$this->_split[$this->_index-1]='';
						if (preg_match('/^\s*/', $this->_split[$this->_index+1], $match)) {
							$this->nl_cnt += (substr_count($match[0], $nl)-1);
							$this->_split[$this->_index+1] = preg_replace('/^\s*/', $nl, $this->_split[$this->_index+1]);	
						}
						$this->_split[$mark_tpl]='';
						$this->_split[$this->_index]='';
					} elseif ($object[0]===16) {
						$this->_split[$this->_index-1]=$object[1];
					} else {
						if ($object[0]&6) $this->_split[$mark_tpl-1] = preg_replace('/(\r\n|\n|\r)[ \t]+$/', '$1', $this->_split[$mark_tpl-1]);
						if ($object[0]&5 and preg_match('/^[ \t]*(\r\n|\n|\r)/', $this->_split[$this->_index+1])) {
							$this->nl_cnt--;
							$this->_split[$this->_index+1] = preg_replace('/^[ \t]*(\r\n|\n|\r)/', '$1$1', $this->_split[$this->_index+1]);
						}
						if ($this->_size_prefix) {
							$object[1] = $this->_size_prefix . $object[1];
							$this->_size_prefix='';
						}
						$this->_split[$this->_index-1]='<?php '.$object[1].'?>';
						$this->_split[$mark_tpl]='';
						$this->_split[$this->_index]='';
					}
				} elseif ($object===-1) {
					$erlist[]=array(htmlspecialchars($this->_split[$this->_index-1]), $this->nl_cnt);
				} else {
					if ($object===-2) $this->report('Error #3', 'unexpected <b>{/}</b>', true, true);
					elseif ($object===-3) $this->report('Error #4', 'unexpected <b>{:}</b>', true, true);
					if (!empty($erlist)) foreach ($erlist as $er) $this->report('Warning #1', '<b>{'.$er[0].'}</b> may be syntax error', true, $er[1]);
					exit;
				}
			}
		}
		if (!empty($this->_control_stack)) {
			$this->report('Error #5', 'template loop or branch is not properly closed by <b>{/}</b>', true);
			exit;
		}
		$source=trim(implode('',$this->_split));
	//
		if (trim($this->postfilter)) $source=$this->_filter($source, 'post');
	//
		$include=$this->_get_function();
		$size_of_top_loop = empty($this->_size_trace[1]) ? '' : $nl.$this->_get_loop_size(1);
		$fp_cpl=@fopen($cpl_path, 'wb');
		if (false===$fp_cpl) {
			$this->report('Error #6', 'cannot write compiled file <b>'.$cpl_path.'</b>. check compile directory or permission');
			exit;
		}
		fwrite($fp_cpl, $cpl_head.$include.$size_of_top_loop.'?>'.$nl.$source);
		fclose($fp_cpl);
		if (!$this->on_ms) @chmod($cpl_path, $this->permission&~0111);
		if ($division) {
			if (!isset($division[$num=count($division)-1]['end'])) $division[$num]['end']=$this->_index;
			for ($i=0,$s=count($division); $i<$s; $i++) {
				$source=($i?$nl:'').trim(implode('', array_slice($this->_split, $division[$i]['start'], $division[$i]['end']-$division[$i]['start'])));
			//
				if (trim($this->postfilter)) $source=$this->_filter($source, 'post');
			//
				$include=$this->_get_function($division[$i]['name']);
				$size_of_top_loop = empty($this->_size_trace[$division[$i]['name']]) ? '' : $nl.$this->_get_loop_size(1, $division[$i]['name']);
				$fp_cpl=fopen($cpl_base.'.'.$division[$i]['name'].'.'.$cpl_ext, 'wb');
				fwrite($fp_cpl, $cpl_head.$include.$size_of_top_loop.'?>'.$nl.$source);
				fclose($fp_cpl);
			}
		}
		return 1;
	}
	function _compile_statement($statement)
	{
		$match=array();
		preg_match('/^(\\\\*)\s*(:\?|[=#@?:\/+])?(.*)$/s', $statement, $match);
		$src=preg_split('/((?:"(?:\\\\"|[^"])*?")|(?:\'(?:\\\\\'|[^\'])*?\'))/', $match[3], -1, PREG_SPLIT_DELIM_CAPTURE);
		for ($i=0;$i<count($src);$i+=2) {
			if (($comment=strpos($src[$i],'//'))!==false) {
				$src[$i]=substr($src[$i], 0, $comment);
				break;
			}
		}
		$src=trim(implode('', array_slice($src, 0, $i+1)));
		if ($match[1]) {
			switch ($match[2]) {
			case '#': case '@': return preg_match('/^[A-Z_a-z\x80-\xff][\w\x80-\xff]*$/',$src) ? array(16, substr($statement,1)) : 0;
			case '/': case ':': return $src ? 0 : array(16, substr($statement,1));
			case '+': return !$src||preg_match('/^[A-Z_a-z\x80-\xff][\w\x80-\xff]*$/',$src) ? array(16, substr($statement,1)) : 0;
			case '' : return $src&&$this->_compile_expression($src,1,1) ? array(16, substr($statement,1)) : 0;
			default : return $src&&$this->_compile_expression($src,1,0) ? array(16, substr($statement,1)) : 0;
			}
		}
		$this->statement=htmlspecialchars($statement);
		switch ($match[2]) { // 1:echo; 2:loop,if,else,elseif,end; 4:include; 8:div;
		case ''  : return ($src and $xpr=$this->_compile_expression($src,0,1)) ? array(1, 'echo '.$xpr) : 0;
		case '=' : return ($src and $xpr=$this->_compile_expression($src,0,0)) ? array(1, 'echo '.$xpr) : 0;
		case ':?': return ($src and $xpr=$this->_compile_expression($src,0,0)) ? array(2, '}elseif('.$xpr.'){') : 0;
		case '+' : return !$src||preg_match('/^[A-Z_a-z\x80-\xff][\w\x80-\xff]*$/', $src) ? array(8, $src) : 0;
		case '#' : return preg_match('/^[A-Z_a-z\x80-\xff][\w\x80-\xff]*$/',$src) ? array(4, '$this->print_($TPL_SCP?$TPL_SCP.".'.$src.'":"'.$src.'",1);') : 0;
		case '?' :
			if (!$src or !$xpr=$this->_compile_expression($src)) return -1;
			$this->_control_stack[]='?';
			return array(2,'if('.$xpr.'){');
		case ':' :
			if ($src) return 0;
			if (empty($this->_control_stack)) return -3;
			if ('@'===array_pop($this->_control_stack)) {
				$this->_close_loop();
				$this->_control_stack[]='?';
				return array(2, '}}else{');
			}
			$this->_control_stack[]='?';
			return array(2, '}else{');
		case '/' :
			if ($src) return 0;
			if (empty($this->_control_stack)) return -2;
			if ('@'===array_pop($this->_control_stack)) {
				$this->_close_loop();
				return array(2,'}}');
			}
			return array(2,'}');
		case '@' :
			if (!preg_match('/^[A-Z_a-z\x80-\xff][\w\x80-\xff]*$/', $src)) return -1;
			$d = ++$this->_loop_depth;
			$this->_control_stack[]='@';
			if ($d>1) {
				if (in_array($src, $this->_loop_stack)) {
					$this->report('Error #7', 'id of nested loop "<b>'.$src.'</b> of <b>{'.$this->statement.'}</b>" cannot be same as parent loop id', true, true);
					exit;
				}
				if (empty($this->_size_trace[$d][$src])) {
					$this->_loop_trace[$d-1]['child_num'] += 3;
					$this->_loop_trace[$d-1]['child_index'][$this->_index-1] = 1;
				}
			}
			if ($d===1 && $this->in_div) $this->_size_trace[$this->in_div][$src] = 1;
			$this->_size_trace[$d][$src] = 1;
			$this->_loop_stack[]=$src;
			$this->_loop_trace[$src]=array('depth'=>$d);
			$this->_loop_trace[$d]  =array('index'=>$this->_index-1, 'foreach'=>0, 'child_num'=>0, 'child_index'=>array());
			$loop_array = $this->_get_loop_array($src, $d);
			return array(2, 'if('.'$TPL_'.$src.'_'.$d.'){for($TPL_I'.$d.'=0,$TPL_R'.$d.'=&'.$loop_array.';$TPL_I'.$d.'<$TPL_'.$src.'_'.$d.';$TPL_I'.$d.'++){');
		}
	}
	function _compile_expression($expression, $escape=0, $noinst=0)
	{
		$prev_is_operand=0;
		$prev_is_func=0;
		$var_state=array(0,'');
		$par_stack=array();
		$this->exp_error=array();
		$this->_foreach=array();
		$this->_for=array();
		$this->_size=array();
		$m=array();
		$func_list=array();
		for ($xpr=''; strlen($expression); $expression=substr($expression, strlen($m[0]))) {
			if (!preg_match('/^
				((?:\.\s*)+)
				|(?:([A-Z_a-z\x80-\xff][\w\x80-\xff]*)\s*(\[|\.|\(|\-\>)?)
				|(?:(\])\s*(\-\>|\.|\[)?)
				|((?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+\-]?\d+)?)
				|((?:"(?:\\\\"|[^"])*")|(?:\'(?:\\\\\'|[^\'])*\'))
				|(===|!==|\+\+|--|\+\.|<<|>>|<=|>=|==|!=|&&|\|\||[,+\-*\/%&^~|<>()!])
				|(\s+)
				|(.+)
			/x', $expression, $m)) return 0;
			if (!empty($m[10])) {
				return 0;
			} elseif ($m[1]) {
				if ($prev_is_operand || $var_state[0]) return 0;
				$prev_is_operand = 1;
				$var_state=array(1,preg_replace('/\s+/','',$m[1]));
			} elseif ($m[2]) {
				if (empty($m[3])) $m[3]='';
				switch ($m[3]) {
				case ''  :
					switch ($var_state[0]) {
					case 0:
						if ($prev_is_operand) return 0;
						$prev_is_operand = 1;
						if (in_array(strtolower($m[2]),$this->key_words) || in_array($m[2], $this->constants)) {
							$xpr.= $m[2];
						} elseif ($m[2][0]==='_') {
							if ($this->safe_mode) $this->exp_error[]=array(4, $m[2]);
							$xpr.= '$GLOBALS["'.substr($m[2],1).'"]';
						} else {
							$xpr.= '$TPL_VAR["'.$m[2].'"]';
						}
						break;
					case 1:
						$xpr.=$this->_compile_array($var_state[1].$m[2], 1);
						break;
					case 2:
						$xpr.= $var_state[1]==='obj' ? $m[2] : '["'.$m[2].'"]';
						break;
					}
					$var_state=array(0,'');
					break;
				case '(' :
					if ($var_state[0]) {
						if ($var_state[1]!=='obj') return 0;
					} else {
						if ($noinst) return 0;
						if ($this->safe_mode && !in_array(strtolower($m[2]),$this->safe_mode_functions)) $this->exp_error[]=array(5, $m[2]);
						$func_list[''][$m[2]]=$this->nl_cnt;
						if ($this->in_div) $func_list[$this->in_div][$m[2]]=$this->nl_cnt;
					}
					$prev_is_operand=0;
					$prev_is_func=1;
					$par_stack[]='f';
					$var_state=array(0,'');
					$xpr.=$m[2].'(';
					break;
				case '[' :
					switch ($var_state[0]) {
					case 0:
						if ($prev_is_operand) return 0;
						$xpr.=$this->_compile_array($m[2]).'[';
						break;
					case 1:
						$xpr.=$this->_compile_array($var_state[1].$m[2]).'[';
						break;
					case 2:
						$xpr.= $var_state[1]==='obj' ? $m[2].'[' : '["'.$m[2].'"][';
						break;
					}
					$par_stack[]='[';
					$prev_is_operand=0;
					$prev_is_func=0;
					$var_state=array(0, '');
					break;
				case '.' :
					switch ($var_state[0]) {
					case 0:
						if ($prev_is_operand) return 0;
						$prev_is_operand=1;
						$var_state=array(1, $m[2].'.');
						break;
					case 1:
						$xpr.=$this->_compile_array($var_state[1].$m[2]);
						$var_state=array(2, '');
						break;
					case 2:
						$xpr.= $var_state[1]==='obj' ? $m[2] : '["'.$m[2].'"]';
						$xpr.=$m[2];
						break;
					}
					break;
				case '->':
					switch ($var_state[0]) {
					case 0:
						if ($prev_is_operand) return 0;
						$prev_is_operand = 1;
						if ($m[2]==='this') {
							if ($this->safe_mode) $this->exp_error[]=array(6, $m[2]);
							$xpr .= '$this->';
						} elseif ($m[2][0]==='_') {
							if ($this->safe_mode) $this->exp_error[]=array(4, $m[2]);
							$xpr .= '$GLOBALS["'.$m[2].'"]->';
						} else {
							$xpr .= '$TPL_VAR["'.$m[2].'"]->';
						}
						break;
					case 1:
						$xpr.=$this->_compile_array($var_state[1].$m[2]).'->';
						break;
					case 2:
						$xpr.=($var_state[1]==='obj' ? $m[2] : '["'.$m[2].'"]').'->';
						break;
					}
					$var_state=array(2,'obj');
					break;
				}
			} elseif ($m[4]) {
				if ($var_state[0] || !$prev_is_operand || empty($par_stack) || array_pop($par_stack)!=='[') return 0;
				if (empty($m[5])) $m[5]='';
				switch ($m[5]) {
				case ''  :
					$xpr.=']';
					break;
				case '->':
					$xpr.=']->';
					$var_state=array(2,'obj');
					break;
				case '.' :
					$xpr.=']';
					$var_state=array(2,'');
					break;
				case '[' :
					$xpr.='][';
					$par_stack[]='[';
					$prev_is_operand=0;
					$prev_is_func=0;
					break;
				}
			} elseif ($m[6]||$m[6]==='0') {
				if ($prev_is_operand) return 0;
				$xpr .= $m[6];
				$prev_is_operand = 1;
			} elseif ($m[7]) {
				if ($prev_is_operand||preg_match('/ [+\-]$/',$xpr)) return 0;
				$xpr=preg_replace('/\+$/','.',$xpr).strtr($m[7],array('``'=>'`','{`'=>'{','`}'=>'}'));
				$prev_is_operand = 1;
			} elseif ($m[8]) {
				if ($var_state[0]) return 0;
				switch ($m[8]) {
				case'++':
				case'--':
					return 0;
				case ',':
					if (!$prev_is_operand || empty($par_stack) || $par_stack[count($par_stack)-1]!=='f') return 0;
					$prev_is_operand=0;
					break;
				case '(':
					if ($prev_is_operand) return 0;
					$par_stack[]='p';
					break;
				case ')':
					if (!$prev_is_operand && !$prev_is_func || empty($par_stack) || array_pop($par_stack)==='[') return 0;
					$prev_is_operand=1;
					break;
				case '!':
				case '~':
					if ($prev_is_operand) return 0;
					break;
				case '-':
					if ($prev_is_operand) $prev_is_operand=0;
					else $m[8]=' -';
					break;
				case '+':
					if (preg_match('/["\']$/', $xpr)) {
						$m[8]='.';
						$prev_is_operand=0;
					} else {
						if ($prev_is_operand) $prev_is_operand=0;
						else $m[8]=' +';
					}
					break;
				case '+.':
					$m[8]='.';
				default	:
					if (!$prev_is_operand) return 0;
					$prev_is_operand=0;
				}
				$xpr .= $m[8];
				$prev_is_func=0;
			} else {
				continue;
			}
		}
		if (!empty($par_stack) || !$prev_is_operand || $var_state[0] || preg_match('/^\d+$/',$xpr)) return 0;
		if ($escape) return 1;
		if (!empty($this->exp_error)) {
			foreach ($this->exp_error as $error) {
				switch ($error[0]) {
				case 3: $this->report('Warning #3', 'loop var <b>'.$error[1].'</b> in <b>{'.$this->statement.'}</b> is not in proper loop',true,true); break;
				case 4: $this->report('Error #9', 'safe mode : global variable <b>'.$error[1].'</b> in <b>{'.$this->statement.'}</b> is not available',true,true); exit;
				case 5: $this->report('Error #10', 'safe mode : function <b>'.$error[1].'()</b> in <b>{'.$this->statement.'}</b> is not registered',true,true); exit;
				case 6: $this->report('Error #11', 'safe mode : <b>this</b> in <b>{'.$this->statement.'}</b> is not available',true,true); exit;
				}
			}
			return 0;
		}
		foreach ($this->_size as $loop_id=>$depth) {
			if ($depth===1 && $this->in_div) $this->_size_trace[$this->in_div][$loop_id] = 1;
			if (empty($this->_size_trace[$depth][$loop_id])) {
				$this->_size_trace[$depth][$loop_id] = array($this->statement, $this->nl_cnt);
				if ($depth>1) {
					$this->_loop_trace[$depth-1]['child_num'] += 3;
					$this->_loop_trace[$depth-1]['child_index'][$this->_index-1] = 1;
				}
			}
		}
		foreach ($this->_foreach as $depth=>$set) {
			$this->_loop_trace[$depth]['foreach'] |= $set;
		}
		foreach ($this->_for as $depth=>$num) {
			$this->_loop_trace[$depth]['child_num'] += $num;
			$this->_loop_trace[$depth]['child_index'][$this->_index-1] = 1;
		}
		if (!empty($func_list)) {
			$this->_set_function('', $func_list);
			if ($this->in_div) $this->_set_function($this->in_div, $func_list);
		}
		return $xpr;
	}
	function _compile_array($subject, $end=0)
	{
		$m=array();
		if (preg_match('/^\.+/', $subject, $m)) {
			$depth=strlen($m[0]);
			if ($this->_loop_depth < $depth) {
				$this->exp_error[]=array(3, $subject);
				return '';
			}
			$id=$this->_loop_stack[$depth-1];
			$var=substr($subject, $depth);
			$el='["'.$var.'"]';
		} else {
			if ($D=strpos($subject,'.')) {
				$id=substr($subject,0,$D);
				$var=substr($subject,$D+1);
				$el='["'.$var.'"]';
				if (in_array($id, $this->_loop_stack)) {
					$depth=$this->_loop_trace[$id]['depth'];
				} elseif ($var==='size_') {
					$depth=$this->_loop_depth+1;
					$this->_size[$id]=$depth;
				} elseif (in_array($var, $this->rsv_words)) {
					$this->exp_error[]=array(3, $subject);
				}
			} else {
				$id=$subject;
				$var='';
				$el='';
				if (in_array($id, $this->_loop_stack)) {
					$depth=$this->_loop_depth;
				}
			}
			if (empty($depth)) {
				if ($id[0]==='_') {
					if ($this->safe_mode) {
						$this->exp_error[]=array(4, $subject);
						return 0;
					}
					if (in_array($id, $this->auto_globals)) return '$'.$id.$el;
					return '$GLOBALS["'.substr($id,1).'"]'.$el;
				}
				return '$TPL_VAR["'.$id.'"]'.$el;
			}
		}
		switch ($var) {
		case 'key_':
			if (!$end) $this->exp_error[]=array(9,$subject);
			elseif (isset($this->_foreach[$id])) $this->_foreach[$id] |= 1;
			else $this->_foreach[$depth] = 1;
			return '$TPL_K'.$depth;
		case 'value_':
			if (isset($this->_foreach[$id])) $this->_foreach[$id] |= 2;
			else $this->_foreach[$depth] = 2;
			return '$TPL_V'.$depth;
		case 'index_':
			if (!$end) $this->exp_error[]=array(9,$subject);
			elseif (isset($this->_foreach[$id])) $this->_foreach[$id] |= 4;
			else $this->_foreach[$depth] = 4;
			return '$TPL_I'.$depth;
		case 'size_':
			if (!$end) $this->exp_error[]=array(9,$subject);
			return '$TPL_'.$id.'_'.$depth;
		default :
			if (isset($this->_for[$depth])) $this->_for[$depth]++;
			else $this->_for[$depth] = 1;
			return '$TPL_R'.$depth.'[$TPL_I'.$depth.']'.$el;
		}
	}
	function _close_loop()
	{
		$loop_id = array_pop($this->_loop_stack);
		$depth   = $this->_loop_depth--;
		$trace   =&$this->_loop_trace[$depth];
		if ($trace['foreach'] & 3) {
			$_key = $trace['foreach']&1 ? '$TPL_K'.$depth.'=>' : '';
			$_idx1='';
			$_idx2='';
			if ($trace['foreach']&4) {
				$_idx1='$TPL_I'.$depth.'=-1;';
				$_idx2='$TPL_I'.$depth.'++;';
			}
			$split = &$this->_split[$trace['index']];
			$split = substr($split, 0, strpos($split,'{')) . '{'.$_idx1.'foreach('.$this->_get_loop_array($loop_id, $depth).' as '.$_key.'$TPL_V'.$depth.'){'.$_idx2.'?>';
		} elseif ($trace['child_num'] > 2) {
			$search  ='$TPL_R'.$depth.'[$TPL_I'.$depth.']';
			$replace ='$TPL_S'.$depth;
			$split =&$this->_split[$trace['index']];
			$split = substr($split, 0, -2) . $replace . '=&' . $search . ';?>';
			$child_indexes = array_keys($trace['child_index']);
			for ($i=0,$s=count($child_indexes); $i<$s; $i++) {
				$split =&$this->_split[$child_indexes[$i]];
				$split = str_replace($search, $replace, $split);
			}
		}
		if (!empty($this->_size_trace[$depth+1])) {
			$split =&$this->_split[$trace['index']];
			$split = substr($split, 0, -2) . $this->_get_loop_size($depth+1) . '?>';
		}
		unset($this->_size_trace[$depth+1], $this->_loop_trace[$depth], $this->_loop_trace[$loop_id]);
	}
	function _get_loop_size($depth, $div='')
	{
		$size  = '';
		$array = $div ? $this->_size_trace[$div] : $this->_size_trace[$depth];
		foreach ($array as $loop_id => $val) {
			if (is_int($depth) && is_array($val)) {
				$this->report('Warning #2', '<b>'.$loop_id.'.size_</b> in <b>{'.$val[0].'}</b> has not corresponding loop', true, $val[1]);
			}
			$loop_array = $this->_get_loop_array($loop_id, $depth);
			$size .= '$TPL_'.$loop_id.'_'.$depth.'=empty('.$loop_array.')||!is_array('.$loop_array.')?0:count('.$loop_array.');';
		}
		return $size;
	}
	function _get_loop_array($loop_id, $depth)
	{
		if ($depth===1) {
			if ($loop_id[0]==='_') {
				if ($this->safe_mode) {
					$this->report('Error #8', 'safe mode : global variable in <b>{'.$this->statement.'}</b> is not available', true, true);
					exit;
				}
				return '$GLOBALS["'.substr($loop_id, 1).'"]';
			}
			return '$TPL_VAR["'.$loop_id.'"]';
		}
		return '$TPL_S'.($depth-1).'["'.$loop_id.'"]';
	}
	function _set_function($divname='', $func_list)
	{
		$prev_list=array_keys($this->func_list[$divname]);
		foreach ($func_list[$divname] as $func => $line) {
			if (!in_array($func, $prev_list)) $this->func_list[$divname][$func]=$line;
		}
	}
	function _get_function($divname='')
	{
		$include=array();
		foreach ($this->func_list[$divname] as $func => $line) {
			if (in_array(strtolower($func), $this->functions)) continue;
			if (!function_exists($func)) {
				$func_file=$this->plugin_fld.'/function.'.$func.'.php';
				$func_path=dirname(__file__).($this->on_ms?'\\':'/').$func_file;
				if (!@file_exists($func_path)) {
					$this->report('Error #12', 'call to undefined function <b>'.$func.'()</b> and cannot find library <b>'.$func_file.'</b>', true, $line);
					exit;
				} elseif (false===include_once $func_path) {
					$this->report('Error #13', 'error in <b>'.$func_file.'</b>', true, $line);
					exit;
				}
				if (!function_exists($func)) {
					$this->report('Error #14', 'cannot find function <b>'.$func.'()</b> in <b>'.$func_file.'</b>', true, $line);
					exit;
				}
			}
			$include[]='"'.$func.'"';
		}
		return $include ? ' $this->include_('.implode(',',$include).');' : '';
	}
	function _filter($source, $type)
	{
		$f_split=preg_split('/(?<!\\\\)\|/', $this->{$type.'filter'});
		$f_sequence=array();
		for ($i=0,$s=count($f_split); $i<$s; $i++) if (trim($f_split[$i])) $f_sequence[]=str_replace('\\|', '|', trim($f_split[$i]));
		if (!empty($f_sequence)) {
			for ($i=0,$s=count($f_sequence); $i<$s; $i++) {
				$f_args=preg_split('/(?<!\\\\)\&/', $f_sequence[$i]);
				for ($j=0,$k=count($f_args); $j<$k; $j++) $f_args[$j]=str_replace('\\&', '&', trim($f_args[$j]));
				$f_name=array_shift($f_args);
				array_unshift($f_args, $source, &$this);
				if (file_exists($f_file=dirname(__file__).'/'.$this->plugin_fld.'/'.$type.'filter.'.$f_name.'.php')) {
					if (false===include_once $f_file) {
						$this->report('Error #15', 'error in filter <b>'.$f_file.'</b>', true);
						exit;
					} elseif (!function_exists($f_name)) {
						$this->report('Error #16', 'filter function <b>'.$f_name.'()</b> is not found in <b>'.$f_file.'</b>');
						exit;
					}
					$source=call_user_func_array($f_name, $f_args);
				} else {
					$this->report('Error #17', 'cannot find filter file <b>'.$f_file.'</b>');
					exit;
				}
			}
		}
		return $source;
	}
	function report($type, $msg, $file=false, $line=false)
	{
		$report = "<br />\n".'<span style="font:12px tahoma,arial;color:#0071DC;background:white">Template_ Compiler '.$type.': '.$msg;
		if ($file) $report.=' in <b>'.(is_string($file)?$file:$this->tpl_path).'</b>';
		if ($line) {
			$line=is_int($line)?$line:$this->nl_cnt;
			foreach ($this->nl_del as $key=>$val) if ($key<=$line) break;
			$report.=' on line <b>'.($line+$val).'</b>';
		}
		echo $report."</span><br />\n";
	}
}
?>